@import {"gplayer.ck", "gship.ck", "gcity.ck", "gobstacle.ck"};
@import {"mouse.ck", "keyboard.ck"};
@import "constants.ck";

//
@import {"ezScore/ezScore.ck", "ezScore/ScorePlayer.ck", "ezScore/ezVoice.ck"}

Constants c;

// ================== variables that represent state of game ==================

// game state
-1 => int current_platform;     // the platform the player is currently standing on (-1 if player isn't on any platform)
false => int play_ending_song;


// ================== read in ezScore ==================

// read in score
ezScore score;
score.setTempo(166);
score.setTimeSig(4, 4);
score.importMIDI("sounds/final.mid");

// update num platforms
6 => int NUM_PARTS;
int NUM_NOTES;
for (int i; i < NUM_PARTS; i++)
{
    score.parts[i].measures[0].notes.size() +=> NUM_NOTES;
}

NUM_NOTES * 2 => int NUM_OBSTACLES;    // *2 is a temp-fix for an ezScore bug (i think), where number of notes isn't the same the number of calls to note_on() functions



// ================== setting up scene ==================
GObstacle obstacles[NUM_OBSTACLES*2];      // size is decided when midi file is read in

// bookkeeping, in case we ungruck an obstacle but later want to put it back
int obstacle_placed[NUM_OBSTACLES];

0 => int current_note_index;

GWindow.title("The Trebled C (3D Edition)");
GWindow.fullscreen();
Keyboard keyboard();
Mouse mouse(0);
spork ~ keyboard.self_update();
spork ~ mouse.self_update();

@(0.0, 0.0, 0.0) => vec3 STARTING_PLAYER_POS;
GPlayer player(keyboard, mouse) --> GG.scene();
player.pos(STARTING_PLAYER_POS);


// start_platform @=> platforms[NUM_NOTES];

// Set up the ship
GShip ship --> GG.scene();
ship.rotateY(-Math.PI/2);
ship.pos(c.STARTING_SHIP_POS);
ship.sca(10);

// Set up the city
4 => int NUM_CITY_BLOCKS;
GCity city_blocks[NUM_CITY_BLOCKS];

for (int i; i < NUM_CITY_BLOCKS; i++)
{
    GCity city_block --> GG.scene();
    city_block.pos(@(-80, -350, -i * 2000));
    city_block.sca(20);
    city_block @=> city_blocks[i];
}


// ================== instruments ==================

Gain g => dac;          // All instruments will connect to g. Must define here so that update_ship_status() can modify sound!
SndBuf buf => dac;

fun place_obstacle(int instrument, ezNote theNote)
{
    GObstacle o(player, instrument);

    (instrument)*40 => float instrument_offset;
    ((theNote.pitch - 70) / 127.0) * 500 => float pitch_offset;

    if (instrument == 4)        // if drums
    {
        ((theNote.pitch - 70) / 127.0) * 100 => float pitch_offset;
    }

    instrument_offset + pitch_offset => float x;
    c.OBSTACLE_HEIGHT => float y;
    player.posWorld().z + player.get_velocity().z => float z;

    o.pos(@(x, y, z));

    o.scaZ(theNote.beats / 0.25);           // 1 beat is a quarter note

    o --> GG.scene();
    o @=> obstacles[current_note_index];
    true => obstacle_placed[current_note_index];

    current_note_index++;
}

// set up starting platform (WALK THE PLANK MATEE)
ezNote note_zero;
place_obstacle(0, note_zero);
obstacles[0].pos(@(0, 0, -20));
obstacles[0].sca(@(1, 1, 6));
obstacles[0].color(@(56, 35, 1) * (1.0/255.0));


class SawOscVoice extends ezVoice
{
    // define sound chain
    8 => n_voices;
    SawOsc oscs[n_voices];
    ADSR envs[n_voices]; 
    Gain g => LPF lpf => outlet;
    8000 => lpf.freq;
    g.gain(.1);
    // rev.mix(.01);
    for(int i; i < n_voices; i++)
    {
        oscs[i] => envs[i] => g;
        envs[i].set(4::ms, 7000::ms, 0.0, 50::ms);
    }

    fun void noteOn(int which, ezNote theNote)
    {
        // <<< "note:", theNote.pitch >>>;
        // <<< "which:", which >>>;
        // <<< "", "" >>>;
        Std.mtof(theNote.pitch) => oscs[which].freq;
        (theNote.velocity / 127.0) => oscs[which].gain;
        envs[which].keyOn();

        place_obstacle(0, theNote);
    }

    fun void noteOff(int which)
    {
        envs[which].keyOff();
    }
}


class TriOscVoice extends ezVoice
{
    // define sound chain
    8 => n_voices;
    TriOsc oscs[n_voices];
    ADSR envs[n_voices]; 
    Gain g => outlet;
    g.gain(.2);
    // rev.mix(.01);
    for(int i; i < n_voices; i++)
    {
        oscs[i] => envs[i] => g;
        envs[i].set(4::ms, 7000::ms, 0.0, 50::ms);
    }

    fun void noteOn(int which, ezNote theNote)
    {
        // <<< "note:", theNote.pitch >>>;
        // <<< "which:", which >>>;
        // <<< "", "" >>>;
        Std.mtof(theNote.pitch) => oscs[which].freq;
        (theNote.velocity / 127.0) / 2 => oscs[which].gain;
        envs[which].keyOn();

        place_obstacle(1, theNote);
    }

    fun void noteOff(int which)
    {
        envs[which].keyOff();
    }
}


class FrencHrnVoice extends ezVoice
{
    // define sound chain
    8 => n_voices;
    FrencHrn hrn[n_voices]; 
    Gain g => NRev rev => dac;
    g.gain(.05);
    rev.mix(.06);
    for(int i; i < n_voices; i++)
    {
        hrn[i] => g;
    }

    fun void noteOn(int which, ezNote theNote)
    {
        <<< "which horn voice:", which >>>;
        Std.mtof(theNote.pitch) => hrn[which].freq;
        (theNote.velocity / 127.0) => hrn[which].noteOn;

        place_obstacle(2, theNote);
    }

    fun void noteOff(int which)
    {
        hrn[which].noteOff(0.5);
    }
}

class ChoirVoice extends ezVoice
{
    // define sound chain
    4 => n_voices;
    SndBuf bufs[n_voices];
    Gain g => outlet;
    g.gain(1);
    for(int i; i < n_voices; i++)
    {
        bufs[i] => g;
    }

    // mapping midi numbers to drum sounds
    55 => int G2;
    57 => int A2;
    58 => int Bb2;
    59 => int B2;

    62 => int D3;
    G2+12 => int G3;
    A2+12 => int A3;
    Bb2+12 => int Bb3;
    B2+12 => int B3;

    D3+12 => int D4;


    fun void noteOn(int which, ezNote theNote)
    {
        theNote.pitch => int pitch;
        // theNote.velocity / 127.0 => float choir_volume;
        1 => float choir_volume;

        // identify which drum sound
        string wav_file_name;
        if (pitch == G2)
        {
            "voices/choir_wavs/G2.wav" => wav_file_name;
        }
        else if (pitch == A2)
        {
            "voices/choir_wavs/A2.wav" => wav_file_name;
        }
        else if (pitch == Bb2)
        {
            "voices/choir_wavs/Bb2.wav" => wav_file_name;
        }
        else if (pitch == B2)
        {
            "voices/choir_wavs/B2.wav" => wav_file_name;
        }
        else if (pitch == D3)
        {
            "voices/choir_wavs/D3.wav" => wav_file_name;
        }
        else if (pitch == G3)
        {
            "voices/choir_wavs/G3.wav" => wav_file_name;
        }
        else if (pitch == A3)
        {
            "voices/choir_wavs/A3.wav" => wav_file_name;
        }
        else if (pitch == Bb3)
        {
            "voices/choir_wavs/Bb3.wav" => wav_file_name;
        }
        else if (pitch == B3)
        {
            "voices/choir_wavs/B3.wav" => wav_file_name;
        }
        else if (pitch == D4)
        {
            "voices/choir_wavs/D4.wav" => wav_file_name;
        }
        else
        {
            <<< "ChoirVoice.ck: error, received invalid midi number for choir:", pitch >>>;
            "voices/choir_wavs/A2.wav" => wav_file_name;
        }

        // play drum sound file
        bufs[which].read(wav_file_name);
        choir_volume => bufs[which].gain;

        place_obstacle(3, theNote);
    }

    fun void noteOff(int which)
    {
        // 0 => bufs[which].gain;
    }
}


class DrumsVoice extends ezVoice
{
    // define sound chain
    6 => n_voices;
    SndBuf bufs[n_voices];
    Gain g => outlet;
    g.gain(0.6);
    for(int i; i < n_voices; i++)
    {
        bufs[i] => g;
    }

    // mapping midi numbers to drum sounds
    54+12 => int HIHAT;    
    41+12 => int RIDE_BELL;
    39+12 => int RIDE_TIP;
    27+12 => int SNARE;
    46+12 => int CRASH_1_EDGE;
    48+12 => int KICK;


    fun void noteOn(int which, ezNote theNote)
    {
        theNote.pitch => int drum_sound;
        // theNote.velocity / 127.0 => float drum_volume;
        1 => float drum_volume;

        // identify which drum sound
        string wav_file_name;
        if (drum_sound == HIHAT)
        {
            "voices/drum_wavs/hihat.wav" => wav_file_name;
        }
        else if (drum_sound == RIDE_BELL)
        {
            "voices/drum_wavs/ride_bell.wav" => wav_file_name;
        }
        else if (drum_sound == RIDE_TIP)
        {
            "voices/drum_wavs/ride_tip.wav" => wav_file_name;
        }
        else if (drum_sound == SNARE)
        {
            "voices/drum_wavs/snare.wav" => wav_file_name;
        }
        else if (drum_sound == KICK)
        {
            "voices/drum_wavs/kick.wav" => wav_file_name;
        }
        else if (drum_sound == CRASH_1_EDGE)
        {
            "voices/drum_wavs/crash_1_edge.wav" => wav_file_name;
        }
        else
        {
            <<< "DrumsVoice.ck: error, received invalid midi number for drums:", drum_sound >>>;
            "voices/drum_wavs/crash_1_edge.wav" => wav_file_name;
        }

        // play drum sound file
        bufs[which].read(wav_file_name);
        drum_volume => bufs[which].gain;

        // place the obstacle
        place_obstacle(4, theNote);
    }

    fun void noteOff(int which)
    {
        // 0 => bufs[which].gain;
    }
}



// ================== updating game state ==================

fun void update_player_normal_force()
{
    true => int disable_player_normal_force;
    for (int i; i < NUM_OBSTACLES; i++)
    {
        // if obstacle i is in contact with the player
        if (obstacles[i].touching_player)
        {
            false => disable_player_normal_force;        // we don't want to turn off normal force
            i => current_platform;                       // we are currently on platform i
            break;                                       // break, since player can only be on platform i (can't be on multiple platforms)
        }
    }

    // if none of the floors are directly under the player
    if (disable_player_normal_force)
    {
        player.set_normal_force(false);     // turn off the normal force
        -1 => current_platform;             // we aren't currently on any platform
    }
}

fun void render_obstacles()
{
    int num_rendered;   // track this for debugging purposes

    for (int i; i < current_note_index; i++)
    {
        if (Math.fabs(player.posWorld().z - obstacles[i].posWorld().z) > c.RENDER_DISTANCE)
        {
            if (obstacles[i].rendered)       // obstacle_placed[i] must be true already if obstacle[i].rendered is true
            {
                obstacles[i] --< GG.scene();
                // null => obstacles[i];
                false => obstacles[i].rendered;
            }
        }

        else
        {
            if (!obstacles[i].rendered && obstacle_placed[i])
            {
                obstacles[i] --> GG.scene();
                true => obstacles[i].rendered;
            }
            
        }

        if (obstacles[i].rendered) num_rendered++;
    }
    // <<< "num rendered obstacles:", num_rendered >>>;
}

fun void clear_obstacles()
{
    // reset visuals
    for (1 => int i; i < current_note_index; i++)
    {
        false => obstacles[i].rendered;
        obstacles[i] --< GG.scene();
    }
    1 => current_note_index;
}

fun void reset_player_and_ship()
{
    @(0, 0, 0) => player.pos;
    @(0, 0, 0) => player.physics_object.position;
    @(0, 0, 0) => player.physics_object.velocity;

    ship --< GG.scene();
    new GShip() @=> ship;
    ship --> GG.scene();
    ship.rotateY(-Math.PI/2);
    ship.pos(c.STARTING_SHIP_POS);
    ship.sca(10);
}
    

fun void reset_game()
{
    clear_obstacles();
    reset_score_player();
    reset_player_and_ship();
    GG.scene().backgroundColor(c.STARTING_BACKGROUND_COLOR);
    
}

GCircle stars[300];
fun void spawn_stars()
{
    for (int i; i < stars.size(); i++)
    {
        GCircle star --> GG.scene();
        @(Math.random2f(-200, 200), Math.random2f(0, 200), Math.random2f(-200, 200)) => vec3 star_offset;
        star.pos(player.posWorld() + star_offset);
    }
    city_blocks[0] --< GG.scene();
}

fun void end_game()
{
    // end chase
    ship.end_chase();

    // play ending song
    true => play_ending_song;

    // disable gravity
    @(0,0,0) => player.physics_object.gravity;

    // spawn stars
    spawn_stars();

    // disable ezscore gain
    g.gain(0);
    
}

fun void check_keypresses()
{
    if (UI.isKeyPressed(UI_Key.R, false))
    {
        reset_game();
    }

    if (UI.isKeyPressed(UI_Key.F, false))
    {
        end_game();
    }
}

// handles updates to player/ship/game status
fun void update_game_status()
{
    player.posWorld().x - ship.posWorld().x => float x_gap;
    player.posWorld().y - ship.posWorld().y => float y_gap;
    Math.fabs(player.posWorld().z - ship.posWorld().z) => float player_dist_from_ship;

    (Math.fabs(x_gap) < 50 && Math.fabs(y_gap) < 70) => int player_in_range;       // 
    
    if (ship.chasing_player)
    {

        // adjust background color (more red as ship approaches)
        1 - ((player_dist_from_ship - c.KILL_DIST) / c.CHASE_DIST) => float r;

        0.15 * r => float shake_delta;
        @(Math.random2f(-shake_delta, shake_delta), Math.random2f(-shake_delta, shake_delta), Math.random2f(-shake_delta, shake_delta)) => vec3 shake_offset;
        player.eye.pos(c.STARTING_EYES_POS + shake_offset);

        // g.gain(r);
        GG.scene().backgroundColor(@(r, 1-r, 1-r));

        <<< "chasing, distance from ship:", player_dist_from_ship >>>;
        <<< "player in range?", player_in_range >>>;
        if (player_dist_from_ship < c.KILL_DIST && player_in_range)
        {
            <<< "SMACK" >>>;
            if (x_gap > 0)  player.launch(@(8000, 0, -8000));
            else player.launch(@(-8000, 0, -8000));
        }
    }
    else
    {
        if (player_dist_from_ship > c.CHASE_DIST)
        {
            <<< "STARTING CHASE" >>>;
            if (!ship.chase_over) ship.begin_chase();
        }
    }

    if (-y_gap > 70)
    {
         <<< "YOU DIED" >>>;
         reset_game();
    }

    check_keypresses();
}




fun void update_state()
{
    update_player_normal_force();
    // update_tunnel_pos();
    render_obstacles();
    update_game_status();
}



// ================== main loop ==================

ScorePlayer sp(score);
fun void reset_score_player()
{
    g.gain(c.STARTING_GAIN);
    sp.pos(0);
    buf.gain(0);
}

fun play_score()
{
    
    5::ms => sp.tick;
    0 => sp.rate;

    SawOscVoice treble => g;
    TriOscVoice treble2 => g;
    FrencHrnVoice accent => g;
    SawOscVoice bass => g;
    ChoirVoice choir => g;
    DrumsVoice drums => g;

    g.gain(c.STARTING_GAIN);

    sp.setVoice(1, treble);
    sp.setVoice(2, treble2);
    sp.setVoice(3, accent);
    sp.setVoice(4, bass);
    sp.setVoice(5, choir);
    sp.setVoice(6, drums);

    while (true)
    {
        c.TICK_DUR => now;
        -player.get_velocity().z / 40 => sp.rate;
    }
}
spork ~ play_score();

GG.scene().backgroundColor(@(0.6, 0.6, 0.6));

while(true)
{
    GG.nextFrame() => now;
    update_state();

    if (play_ending_song)
    {
        buf.read("sounds/piano.wav");
        buf.gain(1);
        false => play_ending_song;
    }

}
